#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "bufmgr.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"
#include "sched.h"
#include "cost.h"


extern "C" {
  double ceil(double x);
  void exit(int);
}

extern int PageSize;

void writeOutputPage(Node*, Op*, File*, int);
void readPage(Node*, Op*, FileScan*, int&, int&);
findBucket(Op*);

// routine to simulate a join operation operating on a disk-resident
// relation

void joinOperator(JoinOp* op)
{

    Node* 	n;
    Msg*	msg;
    CntlMsg	*cMsg;
    DataMsg	*dMsg;
    int		instCnt;
    int		tupleCnt;  // number of tuples processed by operator
    float       tuplesProduced; // running sum of tuples produced
                                // used to model fractional tuples
    int         tupNum;
    int		inTupCnt;
    TIME	startTime; // time operator starts running
    int		splitIdx;	// which split table entry to use	
    int		eofCnt;	  // number of end of stream indicators encountered
    int		tupsPerNetPageOut;
    int		buildTupsPerPage, probeTupsPerPage;


    float	sf;		// local value of operator selectivity factor
    int		splitCnt;
    SplitTable	*splitTable;	
    Mailbox*	inPort;		// mailbox for process to receive messages
    int         bucketNum;      // number of bucket being processed
    int         bNum;           // dummy index variable
    int         pagesReserved;  // actual number of pgs reserved by the BufMgr
    int         hashPages;      // total number of pages available to the hash
                                // table
    int         currentPages;   // total amount of memory used by join till now
    int         e ;             // return code from the buffer manager
    int         scanRC;         // return value from readPage

    int         tupsOnDisk;     // tuples sent to disk
    int         tupsInHash;     // tuples to be inserted in hash table
    float       overflowRatio;  // % of tuples sent to disk
    int         overflowTups;   // number of overflow tuples
    bool        overflow;       // indicates if overflow has occured
    bool        overflowFile;   // indicates if overflow file being processed
    int         outputBucket;   // bucket to which tuple to be sent
    File        *buildOverflowFile, *probeOverflowFile, *currBuildFile,
                *currProbeFile;
    FileScan    *buildFileScan, *probeFileScan;
    int         diskId, buildDisk, probeDisk;
    bool        done;
    BucketRec*  outputArray;

    create("join");

    n = op->node;
    // construct mail box for process to receive messages
    inPort = new Mailbox("opMbox", n, op);

    // initializations 
    startTime = clock;
    tupsPerNetPageOut = NETPKTSIZE/(op->outTupSize + 15);
	// 15 for padding and alignment ???
    buildTupsPerPage = PageSize / op->inTupSizeL;
    probeTupsPerPage = PageSize / op->inTupSizeR;

    sf = op->selFactor;
    if (Flags[TRACE_OP])
      printf("join started on node %d, sf=%f\n",n->id,sf);

    // begin by reserving as many pages as have been allocated to the join
    pagesReserved = n->bufMgr->ReserveBufferPages(op->pagesAllocated, op);
    if (pagesReserved != op->pagesAllocated)
         printf("BIG PROBLEM!!!!!: operator reserved %d pages allocated (%d)\n"
                , pagesReserved, op->pagesAllocated);

    
    // send the operators private port to the query 
    // scheduler that started the operator
    cMsg = CntlMsg::alloc();
    cMsg->opCode = NEWTASK;
    cMsg->intData = n->id;  // kind of a kludge to get the node id returned
    cMsg->send (op->initiator, inPort);
    
    // consume some CPU starting the operator
    n->cpu.use(((float) INITIATE_JOIN) / (Mips * 1000)); // milliseconds
    op->cnt_CPUInstr += INITIATE_JOIN;


    if (((JoinPlan*) (op->query))->numBuckets > 1)
      outputArray = new BucketRec[((JoinPlan*) (op->query))->numBuckets];

    // make local files for each of the extra buckets
    for(bucketNum = 1; bucketNum < ((JoinPlan*) (op->query))->numBuckets; 
        bucketNum++)
    {
        outputArray[bucketNum].numTuples = 0;

        diskId = random(0, DisksPerNode - 1);
	outputArray[bucketNum].buildFile = new File(op->query->transId, 1, &n,
	     1, &(diskId), 0, op->inTupSizeL);

        diskId = random(0, DisksPerNode - 1);
	outputArray[bucketNum].probeFile = new File(op->query->transId, 1, &n,
	     1, &(diskId), 0, op->inTupSizeR);
    }

    //
    // if there are multiple buckets, the join simply loops around after it is 
    //     done with the BUILD and PROBE phase for each bucket
    // for the first bucket the tuples arrive from the child operator directly
    // for the subsequent buckets the tuples are read from local files
    //
    for(bucketNum = 0; bucketNum < ((JoinPlan*) (op->query))->numBuckets; 
        bucketNum++)
    {

       if (Flags[TRACE_OP])
	   printf("Join (node %d) processing bucket %d\n", n->id, bucketNum);

       overflow = FALSE;
       overflowFile = FALSE; 

       // each bucket may produce overflow files so loop over all of them
       do
       {
           // remove pages needed for the local store/selects to find memory
           // available to the hash table
           if (bucketNum == 0)
               if (overflow)
                   hashPages = pagesReserved;
               else
                   hashPages = pagesReserved - 
		               (((JoinPlan*)op->query)->numBuckets - 1);
           else
               hashPages = ((JoinPlan*)op->query)->bucketKMemory ;

	   // if overflow or extra buckets set files from which data to be
	   // read in build and probe phase
           if (overflow) {
               overflowFile = TRUE;  // tuples from overflow file
	       currBuildFile = buildOverflowFile;
	       buildDisk = currBuildFile->storageDisk(0);
	       buildFileScan = new FileScan(currBuildFile, n, buildDisk);
	       currProbeFile = probeOverflowFile;
	       probeDisk = currProbeFile->storageDisk(0);
	       probeFileScan = new FileScan(currProbeFile, n, probeDisk);
	       if (Flags[TRACE_OP])
		   printf("Join (node %d) processing overflow for bucket %d\n",
			  n->id, bucketNum);
	   } else {
	       overflowFile = FALSE;
	       if (bucketNum > 0) {
	           currBuildFile = outputArray[bucketNum].buildFile;
	           buildDisk = currBuildFile->storageDisk(0);
	           buildFileScan = new FileScan(currBuildFile, n, buildDisk);
	           currProbeFile = outputArray[bucketNum].probeFile;
	           probeDisk = currProbeFile->storageDisk(0);
	           probeFileScan = new FileScan(currProbeFile, n, probeDisk);
	       }
	   }

           tupleCnt = 0;
           tupsOnDisk = 0;
           tuplesProduced = 0;
           eofCnt = 0;
           overflowRatio = 0;
           overflow = FALSE;  // reset overflow flag
	   overflowTups = 0;
	   done = FALSE;

           while (!done) {
	       inTupCnt = 0;
	       // if first bucket and no overflow file
               if ((bucketNum == 0) && (overflowFile == FALSE)) {
                    // read pages off the network
	            // block waiting for the next message
	            msg = inPort->recv(); 
	            if (msg->type == CONTROL)  {
		        printf("join on node %d got illegal control msg opcode=%d\n",
			     n->id, cMsg->opCode);
	                (void) CntlMsg::free(cMsg); 
                        exit(NOTOK);
	            } else {
	                // message is another packet of tuples
	                dMsg = (DataMsg*) msg;

//                        if (Flags[TRACE_OP])
//                          printf("join (node %d) rcvd netpage (%d tuples)\n",
//                                 n->id, dMsg->numTuples);

    	                // calculate amount of CPU time consumed copying 
                        // the tuples from the network page to the hash table
   	                if (dMsg->eofFlag == TRUE) {
                             eofCnt++;
                             if (eofCnt == op->inStreamsL) done = TRUE;
                        }

                        // distribute tuples among all the buckets
                        for (tupNum = 0; tupNum < dMsg->numTuples; tupNum++) {
			   outputBucket = findBucket(op);
			   if (outputBucket == 0)
			       inTupCnt++;
			   else {
			       outputArray[outputBucket].numTuples++;

			       // if output page full write to disk
                               if (outputArray[outputBucket].numTuples > buildTupsPerPage) {
				   writeOutputPage(n, op, 
				     outputArray[outputBucket].buildFile,
				     buildTupsPerPage);
				   outputArray[outputBucket].numTuples -= 
				                            buildTupsPerPage;
			       }
			   }
			}
	                (void) DataMsg::free(dMsg); 
                    }   
               } else {
		  // page to be read from local file (overflow/extra buckets)
		  readPage(n, op,  buildFileScan, scanRC, inTupCnt);
		  if (scanRC == END_OF_SCAN) {
		      done = TRUE;
		      e = n->bufMgr->FlushFile(currBuildFile);
		      if (e != OK) 
			printf("Join: buildfile flush returned error %d\n",e);
		      delete buildFileScan;
		      delete currBuildFile;
		  }
	       }

	       // charge cpu for reading the tuples and hashing them
	       instCnt = inTupCnt * (INST_TREAD + INST_BUILD); 
	       n->cpu.use(((float) instCnt) / (Mips * 1000)); // msecs
               op->cnt_CPUInstr += instCnt;

               // if there is an overflow file, send tuples to disk
               if (overflow) {
		      overflowTups += inTupCnt;
		      if (overflowTups > buildTupsPerPage) {
			      writeOutputPage(n, op, buildOverflowFile, 
					      buildTupsPerPage);
			      overflowTups -= buildTupsPerPage;
			      tupsOnDisk += buildTupsPerPage;
		      }
               } else {
                      tupsInHash = inTupCnt;

                      // find tuples and pages in hash table
	              tupleCnt += tupsInHash;
                      currentPages = (int) ceil((tupleCnt * op->inTupSizeL * 
                                      FUDGEFACTOR * 1.0) / PageSize );

//		      printf("current %d hash %d tupsInHash %d tupleCnt %d\n", 
//                           currentPages, hashPages, tupsInHash, tupleCnt);
            
                      // an extra page is added to the capacity of the
                      //   hash table to account for minor differences in the
                      //   estimate of what the memory can hold and what is
                      //   actually sent to the join operator
                      if (currentPages > (hashPages + 1))
		      {
                          // mark spool flag so every next packet can be 
                          // processed properly 
                          overflow = TRUE;
			  if (Flags[TRACE_OP])
			       printf("join-node %d overflow hashPages %d\n", 
				      n->id, hashPages);
			  

			  // make overflowFile
			  diskId = random(0, DisksPerNode - 1);
			  buildOverflowFile = new File(op->query->transId,
			     1, &n, 1, &(diskId), 0, op->inTupSizeL);
		      }
               }
           } // while !done

           if (Flags[TRACE_OP])
           printf("join on node %d has completed build phase HashPages = %d\n",
                  n->id, currentPages);

           if ((bucketNum == 0) && (!overflowFile)) {
              // if this is not an overflow file, send message to controller
              // to indicate that the build phase has completed so the probe
              // phase can start

               cMsg = CntlMsg::alloc();
               cMsg->opCode = TBLSTATUS;
               cMsg->intData = tupleCnt;  // number of tuples built with
               cMsg->queryOp = JOIN;
               cMsg->query = op->query; 
               cMsg->send (op->initiator, inPort);

               // wait for the split table message from controller
               cMsg = (CntlMsg*)inPort->recv(); 
               if (cMsg->opCode != NEWSPLIT) 
               {
                 printf("join on node %d got illegal control msg opcode=%d\n",
	    	        n->id, cMsg->opCode);
               }
               splitTable = cMsg->splitTable; // send made copy of the table
               splitCnt = splitTable->splitCnt;
               cMsg->splitTable = NULL;
               splitTable->tupleSize = op->outTupSize; 
               (void) CntlMsg::free(cMsg); 
           }

           if(overflow) {
		 writeOutputPage(n, op, buildOverflowFile, overflowTups);
		 overflowTups = 0;

                //determine ratio of tuples in hash table to tuples on disk
                //  so the appropriate number of probe tuples can be sent to 
                //  disk
                overflowRatio = (tupsOnDisk * 1.0) / (tupleCnt * 1.0) ;

		// form probe overflow file
		diskId = random(0, DisksPerNode - 1);
		probeOverflowFile = new File(op->query->transId,
			     1, &n, 1, &(diskId), 0, op->inTupSizeR);
           }


	   // write left over tuples for other buckets (only while processing
	   //      bucket 0
	   if ((bucketNum == 0) && (overflowFile == FALSE))
	       for(bNum = 1; bNum < ((JoinPlan*) (op->query))->numBuckets ; 
		   bNum++)  {
		     writeOutputPage(n, op, outputArray[bNum].buildFile,
				 outputArray[bNum].numTuples);
		     outputArray[bNum].numTuples = 0;
	       }


           // begin the probing phase of the join
		   
           tupleCnt = 0;
	   eofCnt = 0;
	   done = FALSE;
	   tuplesProduced = 0;
	   tupsInHash = 0;

           while (!done) {
	       inTupCnt = 0;
               if ((bucketNum == 0) && (overflowFile == FALSE)) {
                    // read pages off the network
	            // block waiting for the next message
	            msg = inPort->recv(); 
	            if (msg->type == CONTROL)  {
		        printf("join on node %d got illegal control msg opcode=%d\n",
			     n->id, cMsg->opCode);
	                (void) CntlMsg::free(cMsg); 
                        exit(NOTOK);
	            } else {
	                // message is another packet of tuples
	                dMsg = (DataMsg*) msg;

//                        if (Flags[TRACE_OP])
//                            printf("join (node %d) rcvd netpage (%d tuples)\n",
//                                   n->id, dMsg->numTuples);

    	                // calculate amount of CPU time consumed copying 
                        // the tuples from the network page to the hash table
   	                if (dMsg->eofFlag == TRUE) {
                             eofCnt++;
                             if (eofCnt == op->inStreamsR) done = TRUE;
                        }

                        // distribute tuples among all the buckets
                        for (tupNum = 0; tupNum < dMsg->numTuples; tupNum++) {
                           outputBucket = findBucket(op);
			   if (outputBucket == 0)
			       inTupCnt++;
			   else {
			       outputArray[outputBucket].numTuples++;

			       // if output page full write to disk
                               if (outputArray[outputBucket].numTuples > probeTupsPerPage) {
				   writeOutputPage(n, op,  
				     outputArray[outputBucket].probeFile,
				     probeTupsPerPage);
			           outputArray[outputBucket].numTuples -=
				                           probeTupsPerPage;
			       }
			   }
                        }
	                (void) DataMsg::free(dMsg); 
		   }   
               } else {
		  // page to be read from local file (overflow/extra buckets)
		  readPage(n, op,  probeFileScan, scanRC, inTupCnt);
		  if (scanRC == END_OF_SCAN) {
		      done = TRUE;
		      e = n->bufMgr->FlushFile(currProbeFile);
		      if (e != OK) 
			printf("Join:probefile flush returned error %d\n",e);
		      delete probeFileScan;
		      delete currProbeFile;
		  }
	       }
	       
	       // Donovans simulator says that this formula is a kludge
	       // as the cost formula should scale with the hash table
	       // chain length

	       instCnt = inTupCnt * (INST_TREAD + INST_PROBE); 
               n->cpu.use(((float) instCnt) / (Mips * 1000)); // msecs
               op->cnt_CPUInstr += instCnt;

	       // if there is an overflow file spool tuples to disk 
	       if (overflow)
               {
		   overflowTups += (int) (inTupCnt * overflowRatio);
		   if (overflowTups > probeTupsPerPage) {
		         writeOutputPage(n, op, probeOverflowFile, 
					 probeTupsPerPage);
			 overflowTups -= probeTupsPerPage;
		   }
		   tupsInHash = inTupCnt - (int) (inTupCnt * overflowRatio); 
	       }
	       else
		   tupsInHash = inTupCnt;
	       
	       tuplesProduced += tupsInHash * sf;
	     
               if (tuplesProduced >= 1.0)
	       {
		    // for each output tuple, hash into split table
		    //   and send netpage if page full
		    for(tupNum = (int) tuplesProduced; tupNum > 0; tupNum--)
		    {
			// now pick a split table entry
			splitIdx = random(0, splitTable->splitCnt - 1);
			  
			splitTable->splitArray[splitIdx].numTuples++;
			if ((splitTable->splitArray[splitIdx].numTuples)
			      >= tupsPerNetPageOut) {
			      // split buffer is full, send off buffer to 
			      // destination
			      splitTable->sendNetBuffer(splitIdx, FALSE, n, 
						        inPort);
			}
		    } // for tupNum ...
	            //update the counts of tuples produced so far
		    tupleCnt += (int) tuplesProduced; 
		    tuplesProduced -= (int) tuplesProduced;
	       }  // if tuplesProduced ...
	   } // while (!done)

	   if (overflow)
	     // write leftover overflow tuples
	     writeOutputPage(n, op, probeOverflowFile, overflowTups);

	   if ((bucketNum == 0) && (overflowFile == FALSE))
	       // write leftover tuples for other buckets
	       for(bNum = 1; bNum < ((JoinPlan*) (op->query))->numBuckets ; 
		   bNum++)  {
		 writeOutputPage(n, op, outputArray[bNum].probeFile,
				 outputArray[bNum].numTuples);
		 outputArray[bNum].numTuples = 0;
	       }

	   if (Flags[TRACE_OP])
              printf("join on node %d has completed probe phase tupCnt = %d\n",
               n->id, tupleCnt);
      } // do ...
      while (overflow);

      if (bucketNum == 0) 
      {
	  // release extra pages needed for bucket 0
          e = n->bufMgr->ReleaseReservedPages(op, op->pagesAllocated -
		                      ((JoinPlan*) op->query)->bucketKMemory);
          if (e != OK) 
             printf("in Join, bufmgr.ReleaseReservedPages returned error\n");
      }

    }

    // exhausted all input streams.  flush all output buffers
    // sends Eofs to the subsequent operator
    splitTable->flushNetBufs(n, inPort);

    //release the pages reserved for this join operator
    e = n->bufMgr->ReleaseBufferPages(op);
    if (e != OK) printf("in Join, bufMgr.releasePage returned error\n");

   
    // all finished.  use some cpu time to terminate the operator 
    // cleanly
    n->cpu.use(((float) TERMINATE_JOIN) / (Mips * 1000)); // msecs
    op->cnt_CPUInstr += TERMINATE_JOIN;

    if (Flags[TRACE_OP])
        printf("join (%d) on node %d finishing time %f\n",
	       identity(), n->id, clock);

    if (((JoinPlan*) op->query)->numBuckets > 0)
	     delete [((JoinPlan*) op->query)->numBuckets] outputArray;

    // delete the split table recieved in the probe phase
    delete splitTable;  

    // send opdone message to the scheduler
    cMsg = CntlMsg::alloc();
    cMsg->opCode = OPDONE;
    cMsg->intData = tupleCnt;  // number of tuples written
    cMsg->queryOp = JOIN;
    cMsg->query = op->query;
    cMsg->send (op->initiator, inPort);

    delete inPort;
    delete op;
}

// add 'numTuples' to 'file'
void writeOutputPage(Node* n, Op* op, File *file, int numTuples)
{
    int e, diskId, pageId;

    diskId = file->storageDisk(0);
    pageId = file->addPage(n,diskId,numTuples);

    // Allocate a page in the bufferpool
    e = n->bufMgr->BucketAllocPage(diskId, pageId, file);
    if (e != OK) printf("Join:BucketAllocPage returned error %d\n",e);
    
    // Then write into the allocated page
    e = n->bufMgr->BucketWritePage(diskId, pageId, op);
    if (e != OK) printf("Join:BucketWritePage returned error %d\n",e);

}

// read page using 'filescan'
void readPage(Node* n, Op* op, FileScan* fileScan, int& scanRC, int& numTuples)
{
    int diskId, pageId, e;

    scanRC = fileScan->nextPage(pageId, diskId, numTuples);
    e = n->bufMgr->BucketReadPage(diskId, pageId, PREFETCH, op);
    if (e != OK) printf("in Select, bufMgr.bucketreadpage returned error\n");
}

// find the bucket to be used to send the tuple. chosen randomly
findBucket(Op* op)
{
    float outputRange;
    int outputBucket;

    // hash into range of bucket values. extra range added to send more
    //     tuples to bucket 0
    outputRange = uniform(0,((JoinPlan*) (op->query))->numBuckets +
	   	                  ((JoinPlan*) (op->query))->weightFactor);
    
    // if value in extra range, send to bucket 0
    if (outputRange > ((JoinPlan*) (op->query))->numBuckets)
        outputBucket =  0;
    else
        outputBucket = (int) outputRange;

    return outputBucket;
}
