#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "query.h"

SelectPlan* SelectPlan::freeSelectPlan = NULL; // no free Select Plans initially
StorePlan* StorePlan::freeStorePlan = NULL; // no free Store Plans initially
TermPlan* TermPlan::freeTermPlan = NULL; // no free Term Plans initially
JoinPlan* JoinPlan::freeJoinPlan = NULL; // no free Join Plans initially

//////////////////////////////////////////////////////////////////////////////

void* SelectPlan::operator new(size_t size)
{
    SelectPlan *newPlan;

    if (freeSelectPlan == NULL) {
	// allocate a new select plan
	newPlan = (SelectPlan *) new char[size];
    } else {
	// reuse an existing select plan
        newPlan = freeSelectPlan;
	freeSelectPlan = freeSelectPlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

void SelectPlan::operator delete(void* oldPlan, size_t)
{
    // put the select plan on the free list
    ((SelectPlan *) oldPlan)->nextPlan = freeSelectPlan;
    freeSelectPlan = (SelectPlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

void* StorePlan::operator new(size_t size)
{
    StorePlan *newPlan;

    if (freeStorePlan == NULL) {
	// allocate a new store plan
	newPlan = (StorePlan *) new char[size];
    } else {
	// reuse an existing store plan
        newPlan = freeStorePlan;
	freeStorePlan = freeStorePlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

void StorePlan::operator delete(void* oldPlan, size_t)
{
    // put the store plan on the free list
    ((StorePlan *) oldPlan)->nextPlan = freeStorePlan;
    freeStorePlan = (StorePlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

void* TermPlan::operator new(size_t size)
{
    TermPlan *newPlan;

    if (freeTermPlan == NULL) {
	// allocate a new term plan
	newPlan = (TermPlan *) new char[size];
    } else {
	// reuse an existing term plan
        newPlan = freeTermPlan;
	freeTermPlan = freeTermPlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

void TermPlan::operator delete(void* oldPlan, size_t)
{
    // put the term plan on the free list
    ((TermPlan *) oldPlan)->nextPlan = freeTermPlan;
    freeTermPlan = (TermPlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

void* JoinPlan::operator new(size_t size)
{
    JoinPlan *newPlan;

    if (freeJoinPlan == NULL) {
	// allocate a new join plan
	newPlan = (JoinPlan *) new char[size];
    } else {
	// reuse an existing join plan
        newPlan = freeJoinPlan;
	freeJoinPlan = freeJoinPlan->nextPlan;
    }
    return newPlan;
};

//////////////////////////////////////////////////////////////////////////////

void JoinPlan::operator delete(void* oldPlan, size_t)
{
    // put the join plan on the free list
    ((JoinPlan *) oldPlan)->nextPlan = freeJoinPlan;
    freeJoinPlan = (JoinPlan *) oldPlan;
};

//////////////////////////////////////////////////////////////////////////////

QueryPlan::QueryPlan()
{
    transId = schedType = -1;
    name = NULL;
    jobClass = NULL;
    statsPlan = NULL;
    startTime = clock;
    outTupSize = noActiveDisk = activeCnt = tupleCnt = 0;
    minMemNeeded = 1.0;  minMemUnits = MINREQ_UNITS; // default is min required
    maxMemNeeded = 1.0;  maxMemUnits = BUFPOOL_UNITS;  // deflt is entire bpool
    cnt_CPUInstr = cnt_bufRef = cnt_bufHit = cnt_bufDirty = cnt_diskRead = 
      cnt_diskCacheHit =0;
};

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::pullStats(QueryPlan* source) 
{
    cnt_bufRef += source->cnt_bufRef;
    cnt_bufHit += source->cnt_bufHit;
    cnt_bufDirty += source->cnt_bufDirty;
    cnt_diskRead += source->cnt_diskRead;
    cnt_diskCacheHit += source->cnt_diskCacheHit;
    cnt_CPUInstr += source->cnt_CPUInstr;
}

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::accumPlanStats() 
{
    statsPlan->bufRefs->record(cnt_bufRef);
    statsPlan->bufHits->record(cnt_bufHit);
    statsPlan->bufDirtys->record(cnt_bufDirty);
    statsPlan->diskReads->record(cnt_diskRead);
    statsPlan->diskCacheHits->record(cnt_diskCacheHit);
    statsPlan->CPUInstr->record(cnt_CPUInstr);
    printf("accum stats, cpu instr = %d\n", cnt_CPUInstr);

}

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::print(int indent) {
    int i;
    char *opName, *plName;

    switch (opCode) {
    case SELECT:  opName = "SELECT"; break;
    case STORE:   opName = "STORE"; break;
    case TERMOP:opName = "TERMINAL"; break;
    case JOIN:    opName = "JOIN"; break;
    case SPOOL:   opName = "SPOOL"; break;
    case DELAY:   opName = "DELAY"; break;
    default:      opName = "???";
    }

    if (name == NULL) 
      plName = "-";          // unnamed plan
    else
      plName = name;
  
    printf("%s(%s): tran %d, sched %d, out sz %d, #act %d, #tups %d",
	   opName, plName, transId, schedType, outTupSize,
	   activeCnt, tupleCnt);
    NewLine;
    printf(" #outtup %d minMem %f(%d) maxMem %f(%d) statsPlan %x", 
	   numOutTuples, minMemNeeded, minMemUnits, maxMemNeeded, 
	   maxMemUnits, statsPlan);
    NewLine;
    printf("sites = { ");
    for (i = 0; i < MAX_NODES; i++) if (activeSites[i]) printf("%d ", i);
    printf("}");
}

//////////////////////////////////////////////////////////////////////////////

void QueryPlan::printName(int indent) {

    switch (opCode) {
    case SELECT:  printf("SELECT"); break;
    case STORE:   printf("STORE"); break;
    case TERMOP:  printf("TERMINAL"); break;
    case JOIN:    printf("JOIN"); break;
    case SPOOL:   printf("SPOOL"); break;
    case DELAY:   printf("DELAY"); break;
    default:      printf("???");
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// find the root of the query tree of this operator is a part
//
QueryPlan* QueryPlan::findRoot()
{
  QueryPlan *qp = this; 
  while (qp->parent != NULL) qp = qp->parent;
  return(qp);
}


//////////////////////////////////////////////////////////////////////////////

SelectPlan::SelectPlan()
{
    opCode = SELECT;
    file = NULL;
    index = NULL;
    inTupSize = -1;
    selFactor = 0.0;
    indexSelectivity = 0.0;
    splitTable = NULL;
    selectType = REMOTE;
};

//////////////////////////////////////////////////////////////////////////////

SelectPlan::~SelectPlan() {
    if (Flags[TRACE_CONTROL])
      printf("running select plan destructor\n"); 
    if (splitTable != NULL) delete splitTable;
};

//////////////////////////////////////////////////////////////////////////////

void SelectPlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    printf(", file %d, sel %f, in sz %d\n", 
	   file->fileNo, selFactor, inTupSize);
}

//////////////////////////////////////////////////////////////////////////////

void SelectPlan::accumStats() {

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

//////////////////////////////////////////////////////////////////////////////

StorePlan::StorePlan()
{
    opCode = STORE;
    file = NULL;
    diskId = 0;
    inStreams = 0;
    inTupSize = -1;
    writeTime = 9999999999.9;
    storeType = REMOTE;
};

//////////////////////////////////////////////////////////////////////////////

void StorePlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    if (storeType == REMOTE)
      printf(", file %d", file->fileNo); 
    printf(", disk %d, #in %d, in sz %d",
	   diskId, inStreams, inTupSize);
    NewSubLine;
    child->print(indent + PRINTINDENT);
}

//////////////////////////////////////////////////////////////////////////////

void StorePlan::accumStats() {

    child->accumStats();           // accum stats into child
    this->pullStats(child);        // pull childs stats into this node

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

/////////////////////////////////////////////////////////////////////////////

TermPlan::TermPlan() 
{
    opCode = TERMOP;
    child = NULL;
    inStreams = 0;
}

/////////////////////////////////////////////////////////////////////////////

void TermPlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    NewSubLine;
    child->print(indent + PRINTINDENT);
}

//////////////////////////////////////////////////////////////////////////////

void TermPlan::accumStats() {

    child->accumStats();           // accum stats into child
    this->pullStats(child);        // pull childs stats into this node

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

//////////////////////////////////////////////////////////////////////////////

JoinPlan::JoinPlan()
{
    opCode = JOIN;
    selFactor = 0.0;
    inStreamsL = 0;
    inStreamsR = 0;
    inTupSizeL = 0;
    inTupSizeR = 0;
    splitTable = NULL;
    numBuckets = 1;
    skewNode = 0;
}

//////////////////////////////////////////////////////////////////////////////

JoinPlan::~JoinPlan() { 
    if (Flags[TRACE_CONTROL])
      printf("running join plan destructor\n");
    if (splitTable != NULL) delete splitTable;
};

//////////////////////////////////////////////////////////////////////////////

void JoinPlan::accumStats() {

    leftChild->accumStats();           // accum stats into left child
    rightChild->accumStats();          // accum stats into right child
    this->pullStats(leftChild);        // pull left child stats into this node
    this->pullStats(rightChild);       // pull right child stats into this node

    // If stats are being maintained for this plan, accumulates them in the
    // appropriate random plan object.

    if (statsPlan != NULL) this->accumPlanStats();
}

//////////////////////////////////////////////////////////////////////////////

void JoinPlan::print(int indent) {
    QueryPlan::print(indent);      // print base class stuff first
    NewLine;
    printf("sel %f, bucks %d, l eofs %d, r eofs %d, l tupsize %d, r tupsize %d, phase %d",
	   selFactor, numBuckets, inStreamsL, inStreamsR, inTupSizeL, 
	   inTupSizeR, phase);
    NewSubLine;
    leftChild->print(indent + PRINTINDENT);
    NewSubLine;
    rightChild->print(indent + PRINTINDENT);
}


