#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "bufmgr.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"
#include "cost.h"

extern "C" {
    void exit(int);
}


extern SkewFactor;

void processScan(Node*, SelectPlan*, Mailbox*);

// declaration for biased random operator KLUDGE FOR SKEW
skewedRandom(int, SelectOp*);

// routine to simulate a selection operation operating on a 
// disk-resident relation

void selectOperator(SelectOp* op)
{
    int         numTuples, tupNum, fullTuples;
    float       tuplesProduced; // running sum of tuples produced
                                // used to model fractional tuples
    Node* 	n;
    Msg         *msg;
    CntlMsg*	cMsg;
    DataMsg*    dMsg;
    int		instCnt;
    int		tupleCnt;  // number of tuples processed by operator
    TIME	startTime; // time operator starts running
    int		tupsPerNetPageOut;
    int		splitIdx;	// which split table entry to use	

    // following four variables are for performance reasons only
    float	sf;		// local value of operator selectivity factor
    Mailbox*	inPort;		// mailbox for process to receive messages
    Mailbox*	scanPort;	// mailbox of associated scan process
    SplitTable*	splitTable;	// local pointer for performance
    SelectPlan* selPlan;
    bool        eofCnt;

    create("select");

    n = op->node;
    // construct mail box for process to receive messages
    inPort = new Mailbox("opMbox", n, op);

    // initializations 
    tupleCnt = 0;     // total number of tuples processed
    startTime = clock;
    tupsPerNetPageOut = NETPKTSIZE/(op->outTupSize + 15);
	// 15 for padding and alignment ???
    sf = op->selFactor;
    selPlan = ((SelectPlan*)op->query);

    splitTable = op->splitTable;
    
    if (Flags[TRACE_OP])
    {
      printf("select (%d) started on node=%d, sf=%f, outtupsize=%d\n", 
	     identity(),n->id, sf,op->outTupSize);
    }

    // begin by sending the operators's private port to the query 
    // scheduler that started the operator
    cMsg = CntlMsg::alloc();
    cMsg->opCode = NEWTASK;
    cMsg->intData = n->id;  // kind of a kludge to get the node id returned
    cMsg->send (op->initiator, inPort);

    // consume some CPU starting the operator
    n->cpu.use(((float) INITIATE_SELECT) / (Mips * 1000)); // milliseconds
    op->cnt_CPUInstr += INITIATE_SELECT;

    tuplesProduced = 0.0;

    // start the associated scan process
    processScan(n, selPlan, inPort);

    eofCnt = FALSE;
    while (!eofCnt) // continue until end of file
    {
         // block waiting for the next message
	 msg = inPort->recv(); 
	 if (msg->type == CONTROL)
	 {  
	     printf("select on node %d got illegal control msg opcode=%d\n",
		    n->id, cMsg->opCode);
	     (void) CntlMsg::free(cMsg); 
	     exit(NOTOK);
	 }
	 else
	 {
	     // message is another packet of tuples
	     dMsg = (DataMsg*) msg;
	     scanPort = dMsg->sourceBox;
	     numTuples = dMsg->numTuples;

//	     if (Flags[TRACE_OP])
//	       printf("select on node %d received netpage (%d tuples)\n",
//		      n->id, numTuples);
	     
	     // calculate amount of CPU time consumed copying the tuples
	     // from the network page to the hash table

             // data page
             //
	     // apply cost of applying predicate only if
             //    a) it is a file scan OR
             //    b) there is a residual predicate (indicated by selFactor not 
             //       being 1)
             //
             if ((((SelectPlan*) op->query)->indexSelectivity == 0) || 
                 (((((SelectPlan*) op->query)->indexSelectivity) != 0) && 
                 (op->selFactor != 1.0)))
	           instCnt = numTuples * (INST_TREAD + INST_PRED);
            else
	          instCnt = numTuples * INST_TREAD;

            n->cpu.use(((float) instCnt) / (Mips * 1000)); //milliseconds
            op->cnt_CPUInstr += instCnt;

	    tuplesProduced += numTuples * sf;

            if (tuplesProduced >= 1.0)
	    {
              fullTuples = (int) tuplesProduced;

              // for each 'whole' tuple produced 
  	      // determine the correct split table index, including
	      // the case that the select is feeding a hybrid join
	      for(tupNum = fullTuples; tupNum > 0; tupNum--)
	      {

		// now pick a split table entry
		// if skew present and this is not a local select
		// use biased random selection
		if (SkewFactor != 1)
		    splitIdx = skewedRandom(splitTable->splitCnt - 1, op);
		else
                    splitIdx = random(0, splitTable->splitCnt - 1);

		splitTable->splitArray[splitIdx].numTuples++;
		if ((splitTable->splitArray[splitIdx].numTuples)
		 	>= tupsPerNetPageOut)
		{
		    // split buffer is full, send off buffer to destination
//		    if (Flags[TRACE_OP])
//		    {
//		      printf("select (%d) on node %d sending netpage (%d tuples) to node %d\n",
//			     identity(), n->id, 
//			     splitTable->splitArray[splitIdx].numTuples,
//		             splitTable->splitArray[splitIdx].destNode->owningNode->id);
//		      printf("splitIdx %d startIdx %d endIdx %d\n",
//			     splitIdx, startIdx, endIdx);
//		    }

		    splitTable->sendNetBuffer(splitIdx, FALSE, n, inPort);
		}
	      } // for tuplesproduced

	      //update the count of tuples produced so far and the tuples 
	      // produced
	      tupleCnt += fullTuples; 
	      tuplesProduced -= fullTuples;

	    } // if tuplesProduced > 1
	    if (dMsg->eofFlag == TRUE) eofCnt = TRUE;
	    (void) DataMsg::free(dMsg);
	    cMsg = CntlMsg::alloc();
	    cMsg->opCode = PAGEDONE;
            cMsg->queryOp = SELECT;
            cMsg->cheapSend (scanPort, inPort);
         } // if (msg->type )... else 
    }

    // exhausted input file.  flush all output buffers
    splitTable->flushNetBufs(n, inPort);

    // all finished.  use some cpu time to terminate the operator cleanly
    n->cpu.use(((float) TERMINATE_SELECT) / (Mips * 1000)); //milliseconds
    op->cnt_CPUInstr += TERMINATE_SELECT;

    if (Flags[TRACE_OP])
      printf("select (%d) on node %d finishing time %f\n", 
             identity(),n->id, clock);

    // block waiting for OPDONE message from the scan operator
    msg = inPort->recv(); 
    if (msg->type != CONTROL)
    {  
        printf("expected CONTROL message in getPortIds. got type=%d\n",
	       msg->type);
        exit(NOTOK);
    }
    else 
    {
        cMsg = (CntlMsg*) msg;
        if (cMsg->opCode != OPDONE)
	{
            printf("unexpected CONTROL message. expected OPDONE.got type = %d\n",
		   cMsg->opCode);
	    exit(NOTOK);
	}
	else
	    (void) CntlMsg::free(cMsg);
    }

    // now send a done message to the initiator
    cMsg = CntlMsg::alloc();
    cMsg->opCode = OPDONE;
    cMsg->intData = tupleCnt;  // number of tuples written
    cMsg->queryOp = SELECT;
    cMsg->query = op->query;
    cMsg->send (op->initiator, inPort);

    delete inPort;  // delete operator's private port
    delete op;	// delete the select operator object

}




// KLUDGE FOR SKEW
// this procedure implements a random selection process which is biased
// towards one particular skew node
skewedRandom(int splitCnt, SelectOp *op)
{
  JoinPlan* joinPlan;
  int splitIdx;

  joinPlan = (JoinPlan*) op->query->parent;

  // assume extra entries for the skew node
  splitIdx = random( 0, splitCnt + (SkewFactor - 1));

  // if splitidx greater than original bounds the entry if for
  // the skew node (set in sched/makesched.c)
  if (splitIdx > splitCnt) splitIdx = joinPlan->skewEntry;

  return(splitIdx);

}

