// class definitions for Select, Store, and Join operators

class SplitTable;
class Node;
class Mailbox;
class QueryPlan;
class File;
class Index;

class Op {
  private:
  public:
    Node*	node;           // node this operator is running on
    int		transId;
    Mailbox*	initiator;	// initiator's mail box
    int		outTupSize;	// in bytes
    int		inTupSize;	// in bytes
    QueryPlan*  query;          // pointer to node in operator tree

    // The following statistics are initially accumulated by the buffer and 
    // disk code on behalf of a particular operator instance.  The operator
    // destructors take responsibility for accumulating their statistics in 
    // their corresponding query plan node before deleting themselves.
    // The accumStats() method on QueryPlan objects recursively accumulates
    // statistics for child plan nodes into the parent.  A terminal then 
    // takes responsibility for accumulating the statistics from the root node
    // of a query plan into the job class the query was issued under.

    int         cnt_bufRef;     // count of buffer references by this op
    int         cnt_bufHit;     // count of buffer hits by this op
    int         cnt_bufDirty;   // count of buffer pages dirtied by this op
    int         cnt_diskRead;   // count of disk reads by this op
    int         cnt_diskCacheHit; // count of disk cache hits by this op
    int         cnt_CPUInstr;   // count of CPU instructions used by this op

    Op(Node*);
    ~Op();
};

class SelectOp: public Op {
  public:
    File*	file;		   // input file descriptor
    Index*      index;             // if != NULL, index to use for scan of file
    float	selFactor;	   // selectivity factor of operation
    float	indexSelectivity;  // index selectivity (used only for iscan)
    SplitTable*	splitTable;	   // split table 

    SelectOp(Node*);
    ~SelectOp();
};
    
class ScanOp: public Op {
  public:
    File*	file;		   // input file descriptor
    Index*      index;             // if != NULL, index to use for scan of file
    float	indexSelectivity;  // index selectivity (used only for iscan)
    SplitTable*	splitTable;	   // split table 

    ScanOp(Node*);
    ~ScanOp();
};
    
class StoreOp: public Op {
  public:
    File*	file;		// output file descriptor
    int		diskId;
    int		sectFree;	// number of free sectors currently allocated
    int		inStreams;	// number of input streams from child op

    StoreOp(Node*);
};

class TermOp: public Op {
  public:
    int		inStreams;	// number of input streams from child op
    TermOp(Node* n) : Op(n) { inStreams = 0; };
};

class JoinOp: public Op {
  public:
    float	selFactor;	// selectivity factor of operation
    int		inTupSizeL;
    int		inTupSizeR;
    int		tupsPerNetPageInL; 
    int		tupsPerNetPageInR; 
    int		tupsPerNetPageOut;
    int		curNetTupFlushed;
    int		inStreamsL;	// number of input streams from left child
    int		inStreamsR;	// number of input streams from right child
    int		eofCnt;		// number of end of streams encountered
    int         pagesAllocated; // total number of pages allocated to the join

    JoinOp(Node*);
    ~JoinOp();
};


// structure to store information about buckets - used only by joins
typedef struct {
    int   numTuples;  // number of tuples on the page
    File  *probeFile, *buildFile;
} BucketRec;
