//  Name:        sched.h
//
//  Title:       Simulator scheduler class definition.
//
//  Change Activity:
//
//     08//91 (MM) - Initial version.
//

// constants to identify left and right local selects
#define LEFT 1
#define RIGHT 0


// return code from createSchedule if the query sent to the global wait queue
//   otherwise OK(1) or NOTOK(0) returned
#define WAIT 2


// fudge factor - used for calculating the size of join hash tables
#define FUDGEFACTOR 1.2



class Operator {
 private:
  static class Operator* freeOperator;
 public:
  QueryPlan*  oprPlan;   // operator in this dependency group
  Operator*  nextOpr;    // pointer to the next operator in the row
  Operator*  nextReq;    // linked list of free operators
  int phase;             // used to the separate the build and the probe 
                         //    operator nodes for a join Plan
  int bucketNum;         // number of the bucket being processed by this 
                         //     operator

  Operator();
  ~Operator();
  void print();
  void* operator new(size_t size);		       // new is overloaded
  void  operator delete(void* oldReq, size_t size);     // delete is overloaded
};

class Schedule {

 private:
  static class Schedule* freeSchedule;
 public:
  Operator*   oprRow;     // list of dependency groups
  Schedule*   nextRow;    // pointer to next dependency group
  Schedule*  nextReq;    // linked list of free operators

  Schedule();
  ~Schedule();
  void print();
  void* operator new(size_t size);		       // new is overloaded
  void  operator delete(void* oldReq, size_t size);     // delete is overloaded
};



//
// query request definition
//
class QueryReq {
  private:
    static class QueryReq* freeQueryReq;
  public:
    Node *node;                    // pointer to scheduling node 
    QueryMsg *reqMsg;              // query request message
    Schedule *schedule;
    Schedule *waitSchedule;
    int minPagesNeeded[MAX_NODES]; // mini number of pages needed at each site 
                                   //      to run the query
    QueryReq *nextReq;      // pointer to next request in queue
    QueryReq *nextFreeReq;

  QueryReq();
  ~QueryReq(){};
  void* operator new(size_t size);		       // new is overloaded
  void  operator delete(void* oldReq, size_t size);     // delete is overloaded
};



//
// query queue class definition
//
class QueryQueue {
  public:

	QueryReq *qhead;
	QueryReq *qtail;
	qtable* queryQueueStat;

    // query queue constructor
    QueryQueue()
        {
            qhead = NULL;
	    qtail = NULL;
	    queryQueueStat = new qtable("query queue wait");
        }


    // query queue destructor
    ~QueryQueue(){};

    // see if queue is empty
    int nonempty()
        {
    	    return(qhead != NULL);
        };

    // insert into query queue
    void enqueue(Node*, QueryMsg*, Schedule*, Schedule*);  

    // remove query from the queue, default = remove head of queue
    QueryReq *dequeue(QueryReq*& query);

    // print queue contents
    void print();
	
    // print queue statistics
    void printStats();

};


// QueryQueue subclass to model queue for queries waiting for
//     memory
class MemoryQueue: public QueryQueue {
  private:
      // check if any site clash present in the two memory arrays
      bool siteClash(int queryPagesNeeded[MAX_NODES],
		     int currentPagesNeeded[MAX_NODES]);

  public:
      MemoryQueue(){};
      ~MemoryQueue(){};

      // check if any of the waiting queries can run given the present number
      //     of pages in BufPages and fork off a query scheduler if they can
      //     the query does not run if it has a site conflict with any of the
      //     waiting queries
      void checkQueue();
      
      // check if any of the queries waiting have a site clash with
      //    the new query whose memory request is in currentPagesNeeded
      bool checkWaiters(int currentPagesNeeded[MAX_NODES]);

      // these methods same as in QueryQueue except that they also
      // update the job class memory wait statistics
      void enqueue(Node*, QueryMsg*, Schedule*, Schedule*);  
      QueryReq *dequeue(QueryReq*& query);
};
