#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "msg.h"
#include "language.h"
#include "db.h"
#include "query.h"
#include "sched.h"
#include "cost.h"
#include "disk.h"
#include <math.h>

extern "C" {
  double floor(double x);
}


// average seek distance sqrt(avg seek distance) = sqrt(1/3) = 1.732
#define AVGSEEKDIST 1.732

void planCost(Schedule*, Schedule*, QueryPlan*);
double findMax(Schedule*);
void cost(QueryPlan*);


double max(double a, double b)
{
if (a > b) return a;
else return b;
}

void planCost(Schedule* schedule, Schedule* waitSched, QueryPlan* root)
{

  Schedule *waitRow = NULL, *schedRow = NULL;
  double maxTime;
  Operator *opr = NULL;
  QueryPlan *plan = NULL;


  // estimate execution times for all the operators in the tree
  cost(root);

  if (schedule != NULL)
    schedRow = schedule->nextRow;

  


  for(waitRow = waitSched; waitRow != NULL; waitRow = waitRow->nextRow)
  {
    maxTime = findMax(waitRow);

    if (schedRow != NULL)
    {
      for(opr = schedRow->oprRow; opr != NULL; opr = opr->nextOpr)
      {
	plan = opr->oprPlan;
	
	if ((plan->opCode == JOIN) && (((JoinPlan*) plan)->estBuildCost != 0))
	  ((JoinPlan*) plan)->estBuildCost = ((JoinPlan*) plan)->estBuildCost + maxTime;
	else
	  plan->execTime = plan->execTime + maxTime;

      }
      schedRow = schedRow->nextRow;
    } 
  }

  if (Flags[TRACE_SCHED]) printf("Estimated plan cost = %f\n", maxTime);

}




double findMax(Schedule* waitRow)
{
  Operator* opr;
  double oprCost, maxCost;
  JoinPlan* joinOpr;

  maxCost = 0;

  for(opr = waitRow->oprRow; opr != NULL; opr = opr->nextOpr)
  {
    switch(opr->oprPlan->opCode)
    {
      case SELECT:
      case STORE :
	oprCost = opr->oprPlan->execTime;
        break;
      case JOIN:
        joinOpr = (JoinPlan*) opr->oprPlan;
        if (joinOpr->estBuildCost != 0)
	{
          oprCost = joinOpr->estBuildCost;
	  joinOpr->estBuildCost = 0;
	}
        else // probe
	  oprCost = opr->oprPlan->execTime;
        break;
    }

    if (oprCost > maxCost)
      maxCost = oprCost;
  }

  return(maxCost);

}


void cost(QueryPlan* oprPlan)
{

  int numJoins;
  double totalTuples, numPages, ioTime, cpuTime, avgIOTime;
  double recievedTuples, localFactor, numMessages, numLeftTuples;
  double cpuInstr, numMessagesSent ;

  SelectPlan* selectOpr;
  StorePlan* storeOpr;
  JoinPlan* joinOpr;


  switch(oprPlan->opCode)
  {
    case SELECT:

      if (Flags[TRACE_SCHED]) printf("SELECT\n");

      selectOpr = (SelectPlan*) oprPlan;

      totalTuples = selectOpr->file->numTuples / selectOpr->noActiveDisk;
      
      selectOpr->numOutTuples = (int) floor(totalTuples * selectOpr->selFactor);

      numPages = totalTuples * selectOpr->inTupSize / PageSize ;

	// selectOpr->file->tuplesPerPage;

      numMessages = totalTuples * selectOpr->selFactor * selectOpr->outTupSize / 
                    NETPKTSIZE ;

      avgIOTime = XferTime + SettleTime + ((float) (MinLatency + DiskRotation))
	/ 2 + SeekFactor * AVGSEEKDIST;
      ioTime = numPages * avgIOTime;


      cpuInstr = totalTuples * (INST_TREAD + INST_PRED) +
	         numMessages * INST_SEND ;

	  // read tuple and apply pred.
         // write selected tuples to output buffers
         //	   totalTuples * selectOpr->selFactor * INST_WRITE;

      cpuTime = cpuInstr / (Mips * 10e3) ;

      selectOpr->execTime = max(ioTime, cpuTime) + 
                            ((INITIATE_SELECT + TERMINATE_SELECT) / (Mips * 10e3)) ;



      if (Flags[TRACE_SCHED])
      {

	printf("\n\n debugging information \n");
	printf("tupsperpage %d\n", selectOpr->file->tuplesPerPage);
	printf("fileTuples %d noActDisks %d selFactor %f\n", 
	  selectOpr->file->numTuples, selectOpr->noActiveDisk, selectOpr->selFactor);
	printf("totalTuples %f outTuples %d numPages %f\n", 
	       totalTuples, selectOpr->numOutTuples, numPages);
	printf(" ioTime %f cpuTime %f execTime %f\n\n", 
	       ioTime, cpuTime, selectOpr->execTime);
      }

      break;

    case STORE:

      if (Flags[TRACE_SCHED]) printf("STORE\n");
      storeOpr = (StorePlan*) oprPlan;

      cost(storeOpr->child);
      
      recievedTuples = storeOpr->child->numOutTuples * storeOpr->inStreams / 
                       storeOpr->noActiveDisk;

      localFactor = 1 / (double) storeOpr->child->noActiveDisk ;

      cpuInstr = ((1 - localFactor) * recievedTuples / NETPKTSIZE) * INST_RCV +
	         recievedTuples * storeOpr->inTupSize * INST_BCOPY ;

      cpuTime = cpuInstr / (Mips * 10e3) ;

      ioTime = (recievedTuples * storeOpr->inTupSize / PageSize) * XferTime;

      storeOpr->execTime = max(ioTime, cpuTime) + 
                           ((INITIATE_STORE + TERMINATE_STORE) / (Mips * 10e3)) ;



      if (Flags[TRACE_SCHED])
      {
	printf("\n\n debugging information \n");
	printf("childTuples %d inStreams %d noactivedisks %d\n", 
	  storeOpr->child->numOutTuples, storeOpr->inStreams, storeOpr->noActiveDisk);
	printf("recievedTuples %f localFactor %f\n", recievedTuples, localFactor);
	printf("cpuTime %f ioTime %f execTime %f\n\n",cpuTime,ioTime,storeOpr->execTime);
      }

      break;

    case JOIN:

      if (Flags[TRACE_SCHED]) printf("JOIN\n");

      joinOpr = (JoinPlan*) oprPlan;

      if (Flags[TRACE_SCHED]) printf("Estimate cost of left child\n");
      cost(joinOpr->leftChild);
      if (Flags[TRACE_SCHED]) printf("Estimate cost of right child\n");
      cost(joinOpr->rightChild);


      numJoins = joinOpr->noActiveDisk;

      recievedTuples = joinOpr->leftChild->numOutTuples * 
	               joinOpr->leftChild->noActiveDisk / numJoins ;

      numMessages = recievedTuples * joinOpr->inTupSizeL / NETPKTSIZE;
      
      localFactor = 1 / (double) joinOpr->leftChild->noActiveDisk ;

      cpuInstr = (1 - localFactor) * numMessages * INST_RCV +
                 recievedTuples * ( INST_TREAD + INST_BUILD);

      cpuTime = cpuInstr / (Mips * 10e3);

      joinOpr->estBuildCost = ( cpuInstr + INITIATE_JOIN ) / (Mips * 10e3) ;



      if (Flags[TRACE_SCHED])
      {
	printf("JOIN- BuildPhaseCost\n\n\n");

	printf("childTuples %d leftChild %d numJoins %d\n", 
	  joinOpr->leftChild->numOutTuples, joinOpr->leftChild->noActiveDisk, numJoins);

	printf("rcvTuples %f\n", recievedTuples);

	printf("inTupsize %d numMessages %f localFactor %f cpuTime %f\n\n", 
	  joinOpr->inTupSizeL, numMessages, localFactor, cpuTime);


	printf("JOIN- ProbePhaseCost\n\n\n");
      }

      recievedTuples = joinOpr->rightChild->numOutTuples * 
	joinOpr->rightChild->noActiveDisk / numJoins;

      numMessages = recievedTuples * joinOpr->rightChild->outTupSize / NETPKTSIZE ;

	localFactor = 1 / (double) joinOpr->rightChild->noActiveDisk ;

	numLeftTuples = joinOpr->leftChild->numOutTuples * 
                        joinOpr->leftChild->noActiveDisk / numJoins ;




      //// ???????? copied from Donovan's simulator.
      // says that size(output) = size(right input) * selectivity factor
      // ok if each right input matches only one left input and if
      // size of output <= size of the larger (right) relation

      joinOpr->numOutTuples = (int) floor(recievedTuples * joinOpr->selFactor);




      numMessagesSent = (1 - 1 / numJoins) * joinOpr->numOutTuples * 
                          joinOpr->outTupSize / NETPKTSIZE;

      cpuInstr = (1 - localFactor) * numMessages * INST_RCV +
	         recievedTuples * (INST_TREAD + INST_PROBE) +
	         numMessagesSent * INST_SEND ;

      cpuTime = cpuInstr / (Mips * 10e3);

      joinOpr->execTime = (cpuInstr + TERMINATE_JOIN) / (Mips * 10e3) ;




      if (Flags[TRACE_SCHED])
      {
        printf("routTuples %d rActDsk %d numJoins %d\n", 
          joinOpr->rightChild->numOutTuples,joinOpr->rightChild->noActiveDisk, numJoins);

        printf("childoutsize %d lOutTuples %d selFactor %f\n", 
          joinOpr->rightChild->outTupSize, joinOpr->leftChild->numOutTuples, 
	  joinOpr->selFactor);

        printf("numMessages %f localFactor %f recievedTuples %f\n", 
          numMessages, localFactor, recievedTuples);

        printf("cpuTime %f numLeftTuples %f numOutTuples %d\n\n", 
          cpuTime, numLeftTuples, joinOpr->numOutTuples);
      }

      break;
    }

    return;
}

 


/*
  cost for the operators
    assuming uniform distribution of tuples across all the partitions


Q-  to be more sophisticated in estimating io interference, we can look at all
  the operators running in parallel and see if there are any operators running
  on overlapping disks. from this we can make an estimate of the io time for
  each operator.


*/

