#include "dbsim.h"
#include "object.h"
#include "msg.h"
#include "mailbox.h"
#include "node.h"
#include "language.h"
#include "db.h"
#include "terminal.h"
#include "split.h"
#include "query.h"

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        termproc.c
//
//  Function:    Terminal process.  Thinks and submits queries according to
//               its job class description.  Gathers and prints response time 
//               statistics.
//                 
//  Change Activity:
//
//     10/91 (KPB/DJD) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

extern Mailbox* GlobalSchedPort;
extern event    SimDone;
void            cleanUp(QueryPlan*);
extern "C" {
  void current_stats();
  void exit(int code);
}

void TermProc(Terminal *t) {
  float         thinkTime;
  TIME	        startTime, endTime;
  char*         planName;
  bool          reset = FALSE;          // plan sequence has reset to beginning

  distObjType   objType;
  RandomPlan*   rp;
  QueryPlan*	qp;
  QueryMsg*	reqMsg;
  QueryMsg*	replyMsg;
  JobClass*     jc = t->jobClass;      // to shorten ptr expressions
  
  create("terminal");                  // define a CSIM process
  if (Flags[TRACE_TERM] || Flags[TRACE_CONTROL]) 
    printf("Terminal process %d, class %s started\n",t->id, t->jobClass->name);

  while(SimRunning && !reset) {        // loop until timer pop or no more plans
    thinkTime = t->thinksecs->choose() * 1000.0; // pick it & convert sec->msec
    hold(thinkTime);                   // think

    planName = (char*) t->plans->choose(objType, reset); // pick a plan name
    if (reset) break;                  // end of plan list for this terminal

    rp = (RandomPlan*) PlanSyms.lookup(planName); // find random plan object
    if (rp == NULL) {
      printf("ERROR:  terminal %d found an undefined plan: %s\n",
	     t->id, planName);
      exit(4);
    }

    qp = rp->choose(NULL, jc);         // stamp out a specific plan instance

    // Croaks if the root operator of a plan is not a STORE or TERMSEND.  
    // This must be the case, since the result tuples from a query must go
    // somewhere.  KLUDGE !!!!!!!!  Future update/insert ops could also be
    // valid roots!

    if ((qp->opCode != STORE) && (qp->opCode != TERMOP))
        printf("\nERROR: Query rejected - root of query plan not a store/termsend\n\n");
    else {


        reqMsg = new QueryMsg();           // get a query message
        reqMsg->query = qp;                // stuff plan in message
    
	// Sends the query to the global query scheduler

	if (Flags[TRACE_CONTROL]) current_stats();  // dump current statistics
	
	if(Flags[TRACE_TERM]) {
	    printf("terminal %d sending query to global scheduler\n ",t->id);
	    qp->print(0);
	    printf("\n\n");
	}
    
	startTime = clock;
	reqMsg->send(GlobalSchedPort, (Mailbox*) t->port);
    
	// Waits for a reply
      
	replyMsg = (QueryMsg*) t->port->recv();
	endTime = clock;
    
	// Accumulates query stats into job class statistics in CSIM tables

	qp->accumStats();                   // roll up plan stats to root node
        jc->bufRefs->record(qp->cnt_bufRef);
	jc->bufHits->record(qp->cnt_bufHit);
	jc->bufDirtys->record(qp->cnt_bufDirty);
	jc->diskReads->record(qp->cnt_diskRead);
	jc->diskCacheHits->record(qp->cnt_diskCacheHit);
	jc->CPUInstr->record(qp->cnt_CPUInstr);
	jc->respTimes->record(endTime-startTime);
	jc->completions++;                  // bump completions for throughput

	// If root node of query is flagged for plan based statistics, record
        // response times on a plan basis, as well.

	if (qp->statsPlan != NULL)
	  qp->statsPlan->respTimes->record(endTime-startTime);

	if(Flags[TRACE_TERM])
	{
	    printf("\nTerminal %d, query executed in %5.2f ms.", 
		   t->id, endTime-startTime);
	    printf(" started at %f\n\n", startTime);
	}
				     
	delete replyMsg;


	if (Flags[TRACE_CONTROL]) current_stats();  // dump current statistics

    }

    cleanUp(qp);                         // clean up the query

  }
    
  // If this is the last active terminal in the system, or this is the last
  // active terminal of a "primary" job class, then shuts down the simulation.

  if (--ActiveTerms == 0 || --(jc->stopCount) == 0) {
    SimRunning = FALSE;
    SimDone.set(); 
  }
}
     
///////////////////////////////////////////////////////////////////////////////

void cleanUp(QueryPlan* node) { // recursive proc to clean up nodes of a query
  StorePlan*	storeOp;
  JoinPlan*	joinOp;
  SelectPlan*	selectOp;
  TermPlan*	termOp;
  
  switch (node->opCode)   {
  case STORE:
    storeOp = (StorePlan *) node;
    cleanUp(storeOp->child);
    delete storeOp;
    break;
  case TERMOP:
    termOp = (TermPlan *) node;
    cleanUp(termOp->child);
    delete termOp;
    break;
    
  case SELECT:
    selectOp = (SelectPlan *) node;
    delete selectOp;
    break;
    
  case JOIN:
    joinOp = (JoinPlan *) node;
    cleanUp(joinOp->leftChild);
    cleanUp(joinOp->rightChild);
    delete joinOp;
    break;

  default:
    printf("Termproc.cleanUp, unknown opcode=%d\n",node->opCode);
  }
}
